-module(srv).
-behaviour(gen_server).

-export([start/1, stop/0, dump/0]).
-export([connect/1, disconnect/0, send/1, join/1]).

% Callback functions
-export([init/1, handle_cast/2, handle_call/3, handle_info/2,
        terminate/2, code_change/3]).

-define(CLIENT,cli).
-define(SERVERNAME, {global, chatsrv}).

% -------------------- Interface functions ---------------------------
start(Max) ->
    gen_server:start(?SERVERNAME, srv, Max, []).

stop() ->
    gen_server:cast(?SERVERNAME, stop).

dump() ->
    gen_server:call(?SERVERNAME, dump).

% -------------------- Client interface -----------------------------
connect(Nick) ->
    gen_server:call(?SERVERNAME, {connect, Nick}, 1000).

disconnect() ->
    gen_server:cast(?SERVERNAME, {disconnect, self()}).

send(Text) ->
    gen_server:cast(?SERVERNAME, {text, self(), Text}).

join(Group) ->
    gen_server:cast(?SERVERNAME, {join, self(), Group}).

% -------------------- Implementation functions ----------------------
-record(state, {users, max}).
-record(user, {nick, pid, group = "Global"}).

init(Max) ->
    process_flag(trap_exit, true),
    {ok, #state{users=ets:new(users, [{keypos, #user.pid}]), max=Max}}.


handle_cast(stop, S) ->
    {stop, normal, S};

handle_cast({disconnect, Pid}, S=#state{users=Users}) ->
    ets:delete(Users, Pid),
    {noreply, S};

handle_cast({text, From, Text}, S=#state{users=Users}) ->
    [#user{nick=Nick, group=Group}] = ets:lookup(Users, From),
    lists:foreach(
        fun([Pid]) ->
            ?CLIENT:send(Pid, Nick ++ ": " ++ Text)
        end,
        ets:match(Users, #user{group=Group,pid='$1',_='_'})),
    {noreply, S};

handle_cast({join, Pid, Group}, S=#state{users=Users}) ->
    [User] = ets:lookup(Users, Pid),
    ets:insert(Users, User#user{group = Group}),
    {noreply, S}.

handle_call(dump, _From, S=#state{users=Users}) ->
   {reply,
    {S#state.max, ets:match(Users, #user{nick='$1', group='$2', _='_'})},
    S};

handle_call({connect, Nick}, {Pid, _Tag}, S=#state{users=Users, max=Max}) ->
    if
        length(Users) >= Max ->
            {reply, deny, S};
        true ->
            link(Pid),
            ets:insert(Users, #user{nick=Nick, pid=Pid}),
            {reply, ok, S}
    end.


handle_info({'EXIT', From, _}, S=#state{users=Users}) ->
    ets:delete(Users, From),
    {noreply, S}.

terminate(_Reason, #state{users=Users}) ->
    ets:delete(Users).

code_change(_, S, _) ->
    {ok, S}.
