-module(srv).
-behaviour(gen_server).

-export([start/1, stop/0, dump/0]).
-export([connect/1, disconnect/0, send/1, join/1]).

% Callback functions
-export([init/1, handle_cast/2, handle_call/3, handle_info/2,
        terminate/2, code_change/3]).

-include("srv.hrl").

% -------------------- Interface functions ---------------------------
start(Max) ->
    gen_server:start_link(?SERVERNAME, srv, Max, []).

stop() ->
    gen_server:cast(?SERVERNAME, stop).

dump() ->
    gen_server:call(?SERVERNAME, dump).

% -------------------- Client interface -----------------------------
connect(Nick) ->
    gen_server:call(?SERVERNAME, {connect, Nick}, 1000).

disconnect() ->
    gen_server:cast(?SERVERNAME, {disconnect, self()}).

send(Text) ->
    gen_server:cast(?SERVERNAME, {text, self(), Text}).

join(Group) ->
    gen_server:cast(?SERVERNAME, {join, self(), Group}).

% -------------------- Implementation functions ----------------------
-record(state, {users, groups, max}).

init(Max) ->
    error_logger:info_report("Chat server starting."),
    process_flag(trap_exit, true),
    {ok, #state{users=ets:new(users, [{keypos, #user.pid}]),
                groups=ets:new(groups, []),
                max=Max}}.

handle_cast(stop, S) ->
    {stop, normal, S};

handle_cast({disconnect, Pid}, S=#state{users=Users}) ->
    ets:delete(Users, Pid),
    {noreply, S};

handle_cast({text, From, Text}, S=#state{users=Users, groups=Groups}) ->
    [#user{nick=Nick, group=Group}] = ets:lookup(Users, From),
    [{_, Grp}] = ets:lookup(Groups, Group),
    grp_srv:text(Grp, Nick, Text),
    {noreply, S};

handle_cast({join, Pid, Group}, S=#state{users=Users, groups=Groups}) ->
    [User] = ets:lookup(Users, Pid),
    join_grp(Users, Groups, User, Group),
    {noreply, S}.

handle_call(dump, _From, S=#state{users=Users, groups=Groups}) ->
   {reply,
    {S#state.max,
     ets:match(Users, #user{nick='$1', group='$2', _='_'}),
     ets:match(Groups, {'$1', '$2'})
    },
    S};

handle_call({connect, Nick}, {Pid, _Tag}, S=#state{users=Users,
                                                   groups=Groups,
                                                   max=Max}) ->
    if
        length(Users) >= Max ->
            {reply, deny, S};
        true ->
            link(Pid),
            User = #user{nick=Nick, pid=Pid},
            join_grp(Users, Groups, User, "Global"),
            {reply, ok, S}
    end.

join_grp(Users, Groups, User, Group) ->
    case ets:lookup(Groups, Group) of
        [] ->
            ets:insert(Groups, {Group, grp_sup:child(Users, Group)});
        _ -> ok
    end,
    ets:insert(Users, User#user{group = Group}).

handle_info({'EXIT', From, _}, S=#state{users=Users}) ->
    ets:delete(Users, From),
    {noreply, S}.

terminate(_Reason, #state{users=Users}) ->
    error_logger:info_report("Chat server stopping."),
    ets:delete(Users).

code_change(_, S, _) ->
    {ok, S}.
